use std::collections::{HashMap, HashSet};

use crate::parsing::{Message, Transition};
use btrun::{Activate, End};

use super::Protocol;
use proc_macro2::{Ident, TokenStream};
use quote::{format_ident, quote, ToTokens};

impl ToTokens for Protocol {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        tokens.extend(self.generate_message_enum());
        tokens.extend(self.generate_state_traits());
    }
}

impl Protocol {
    fn generate_message_enum(&self) -> TokenStream {
        let mut msgs: HashSet<&Message> = HashSet::new();
        for transition in self.transitions.iter() {
            // We only need to insert received messages because every sent message has a
            // corresponding receiver thanks to the validator.
            if let Some(msg) = &transition.in_msg {
                msgs.insert(msg);
            }
        }
        let variants = msgs.iter().map(|msg| msg.variant());
        let msg_types = msgs.iter().map(|msg| msg.type_tokens());
        let enum_name = format_ident!("{}Msgs", self.name_def.name);
        quote! {
            pub enum #enum_name {
                #( #variants(#msg_types) ),*
            }
        }
    }

    fn generate_state_traits(&self) -> TokenStream {
        let mut traits: HashMap<&Ident, Vec<&Transition>> = HashMap::new();
        for transition in self.transitions.iter() {
            let vec = traits
                .entry(&transition.in_state.state_trait)
                .or_insert_with(Vec::new);
            vec.push(transition);
        }
        let mut tokens = TokenStream::new();
        for (trait_ident, transitions) in traits {
            quote! {
                pub trait #trait_ident {
                    #( #transitions )*
                }
            }
            .to_tokens(&mut tokens);
        }
        tokens
    }
}

impl Message {
    /// Returns the tokens which represent the type of this message.
    fn type_tokens(&self) -> TokenStream {
        // We generate a fully-qualified path to the Activate message so that it doesn't need to be
        // imported for every protocol definition.
        let msg_type = if self.msg_type == Activate::ident() {
            quote! { ::btrun::Activate }
        } else {
            let msg_type = &self.msg_type;
            quote! { #msg_type }
        };
        if self.is_reply {
            quote! {
                <#msg_type as ::btrun::CallMsg>::Reply
            }
        } else {
            quote! {
                #msg_type
            }
        }
    }
}

impl ToTokens for Transition {
    fn to_tokens(&self, tokens: &mut TokenStream) {
        let transition = if let Some(msg) = &self.in_msg {
            self.generate_message_handler(msg)
        } else {
            self.generate_sender_method()
        };
        tokens.extend(transition);
    }
}

impl Transition {
    fn generate_message_handler(&self, msg: &Message) -> TokenStream {
        let msg_type = if msg.msg_type == Activate::ident() {
            quote! { ::btrun::Activate }
        } else {
            msg.msg_type.to_token_stream()
        };
        let method_ident = format_ident!("handle_{}", msg.msg_type.to_lowercase());
        let output_pairs: Vec<_> = self
            .out_states
            .as_ref()
            .iter()
            .map(|state| {
                let state_trait = &state.state_trait;
                if state_trait == End::ident() {
                    (quote! {}, quote! { ::btrun::End })
                } else {
                    let assoc_type = format_ident!("{}Output{}", msg.variant(), state_trait);
                    let output_decl = quote! { type #assoc_type: #state_trait; };
                    let output_type = quote! { Self::#assoc_type };
                    (output_decl, output_type)
                }
            })
            .collect();
        let output_decls = output_pairs.iter().map(|(decl, _)| decl);
        let output_types = output_pairs.iter().map(|(_, output_type)| output_type);
        let future_name = format_ident!("Handle{}Fut", msg.variant());
        quote! {
            #( #output_decls )*
            type #future_name: ::std::future::Future<Output = Result<( #( #output_types ),* )>>;
            fn #method_ident(self, msg: #msg_type) -> Self::#future_name;
        }
    }

    fn generate_sender_method(&self) -> TokenStream {
        quote! {}
    }
}

trait ToLowercase {
    fn to_lowercase(&self) -> String;
}

impl ToLowercase for Ident {
    fn to_lowercase(&self) -> String {
        let mut buf = self.to_string();
        buf.make_ascii_lowercase();
        buf
    }
}
